from flask import Flask, Response, make_response, request
from pydantic import AnyUrl, BaseModel, Field
from werkzeug.exceptions import HTTPException


class Link(BaseModel):
    href: str
    meta: dict | None = None


class JsonApi(BaseModel):
    version: str | None = None
    meta: dict | None = None


class Error(BaseModel):
    id: str | None = None
    status: str | None = None
    code: str | None = None
    title: str | None = None
    detail: str | None = None
    source: dict | None = None
    meta: dict | None = None


class ResourceIdentifier(BaseModel):
    type: str
    id: str
    meta: dict | None = None


class Relationship(BaseModel):
    data: ResourceIdentifier | list[ResourceIdentifier] | None = None
    links: dict[str, AnyUrl | Link | None] | None = None
    meta: dict | None = None


class Resource(BaseModel):
    type: str
    id: str
    attributes: dict = Field(default_factory=dict)
    relationships: dict[str, Relationship] | None = None
    links: dict[str, AnyUrl | Link | None] | None = None
    meta: dict | None = None

    def flatten(self):
        return {"id": self.id, **self.attributes}


class TopLevel(BaseModel):
    data: Resource | list[Resource] | None = None
    errors: list[Error] | None = None
    meta: dict | None = None
    jsonapi: JsonApi | None = None
    links: dict[str, AnyUrl | Link | None] | None = None
    included: list[Resource] | None = None


class JsonApiApp:
    """Flask's extension implementing JSON:API specification."""

    def __init__(self, app: Flask | None = None):
        self.app: Flask | None = None

        if app:
            self.init_app(app)

    def init_app(self, app: Flask):
        self.app = app

        handle_exception = app.error_handler_spec[None][None].get(Exception)
        app.after_request(self._change_content_type)

        def _handle_exception(e: Exception) -> Response:
            status_code: str
            title: str
            detail: str

            if handle_exception:
                try:
                    resp = handle_exception(e)
                    if not isinstance(resp, Response):
                        resp = make_response(resp)
                    status_code = str(resp.status_code)
                    title = str(e)
                    detail = str(resp.data)
                except Exception as ex:
                    e = ex

            if isinstance(e, HTTPException):
                status_code = str(e.code or 500)
                title = e.name
                detail = e.description or str(e)
            else:
                status_code = "500"
                title = type(e).__name__
                detail = str(e)

            error = Error(id=status_code, title=title, detail=detail, status=status_code)
            toplevel = TopLevel(errors=[error]).model_dump_json(exclude_none=True)
            return make_response(toplevel, status_code, {"Content-Type": "application/vnd.api+json"})

        app.error_handler_spec[None][None][Exception] = _handle_exception

    def _change_content_type(self, resp: Response):
        if resp.content_type != "application/vnd.api+json" and resp.status_code >= 400:
            status_code: str = str(resp.status_code)
            error = Error(id=status_code, title=resp.data, detail=resp.data, status=status_code)
            toplevel = TopLevel(errors=[error]).model_dump_json(exclude_none=True)
            return make_response(toplevel, resp.status_code, {"Content-Type": "application/vnd.api+json"})

        if "application/vnd.api+json" not in request.headers.getlist("accept"):
            return resp

        resp.content_type = "application/vnd.api+json"
        return resp



from .client import JapydClient, TopLevelArray, TopLevelSingle
from .dotnet import JsonApiBaseModel, JsonApiBodyModel, JsonApiQueryModel
from .jsonapi import (
    Error,
    JsonApiApp,
    Link,
    Relationship,
    Resource,
    ResourceIdentifier,
    TopLevel,
)

__all__ = [
    "TopLevel",
    "Resource",
    "ResourceIdentifier",
    "Relationship",
    "Link",
    "Error",
    "JsonApiApp",
    "JsonApiBaseModel",
    "JsonApiQueryModel",
    "JsonApiBodyModel",
    "JapydClient",
    "TopLevelSingle",
    "TopLevelArray",
]



from __future__ import annotations

import re
import typing as t
from enum import StrEnum
from functools import cached_property
from http import HTTPStatus

from flask import Response, request
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
    field_validator,
    model_validator,
)
from werkzeug.exceptions import NotFound, UnprocessableEntity

from .jsonapi import Error, Resource, TopLevel
from .models import JsonApiBaseModel
from .utils import to_string_or_numeric


class Oper(StrEnum):
    NOT = "not"
    OR = "or"
    AND = "and"

    EQUALS = "equals"
    LESS_THAN = "lessThan"
    LESS_OR_EQUAL = "lessOrEqual"
    GREATER_THAN = "greaterThan"
    GREATER_OR_EQUAL = "greaterOrEqual"

    CONTAINS = "contains"
    STARTS_WITH = "startsWith"
    ENDS_WITH = "endsWith"

    ANY = "any"  # Equals one value from set
    HAS = "has"  # Collection contains items


FIELDS_REGEXP = re.compile(r"fields\[(.*)]")

NUMERIC_REGEXP = r"[-+]?[0-9]*.?[0-9]+([eE][-+]?[0-9]+)?"
IDENTIFIER_REGEXP = r"([A-Za-z_][A-Za-z0-9_\.]*)"
QUOTED_REGEXP = r"('[^']|'')*'"
VALUE_REGEXP = rf"(?P<value>'([^']|'')*')|(?P<numeric>{NUMERIC_REGEXP})|(?P<other_attr>{IDENTIFIER_REGEXP})"
OPER_REGEXP = (
    rf"({Oper.EQUALS}|{Oper.LESS_THAN}"
    rf"|{Oper.LESS_OR_EQUAL}|{Oper.GREATER_THAN}|{Oper.GREATER_OR_EQUAL}"
    rf"|{Oper.CONTAINS}|{Oper.STARTS_WITH}|{Oper.ENDS_WITH})"
)
LIST_OPER_REGEXP = rf"({Oper.ANY}|{Oper.HAS})"
LIST_VALUES_REGEXP = rf"(('([^']|'')*')|{NUMERIC_REGEXP})"
FILTER_REGEXP = (
    rf"(?P<oper>{OPER_REGEXP})\s*[(]\s*(?P<attr>{IDENTIFIER_REGEXP})\s*,\s*({VALUE_REGEXP})\s*[)]"
    rf"|(?P<list_oper>{LIST_OPER_REGEXP})\s*[(]\s*(?P<list_attr>{IDENTIFIER_REGEXP})\s*,"
    rf"\s*(?P<values>{LIST_VALUES_REGEXP}(\s*[,]\s*{LIST_VALUES_REGEXP}\s*)*)\s*[)]"
)
COMPILED_FILTER_REGEXP = re.compile(rf"^\s*{FILTER_REGEXP}\s*$")

NOT_REGEXP = rf"\s*{Oper.NOT}\s*[(]\s*(?P<filter>.*)\s*[)]\s*"
COMPILED_NOT_REGEXP = re.compile(NOT_REGEXP)

OR_AND_REGEXP = rf"\s*({Oper.OR}|{Oper.AND})\s*[(]\s*(?P<filter1>.*)\s*,\s*(?P<filter2>.*)\s*[)]\s*"
COMPILED_OR_AND_REGEXP = re.compile(OR_AND_REGEXP)


class JsonApiPagination(BaseModel):
    number: int
    size: int
    total: int | None = None


class JsonApiQueryFilter(BaseModel):
    model_config = ConfigDict(str_strip_whitespace=True)

    oper: Oper
    subfilter: tuple[JsonApiQueryFilter, JsonApiQueryFilter] | JsonApiQueryFilter | None = None
    attr: str | None = None
    value: str | float | int | list | None = None
    other_attr: str | None = None

    @model_validator(mode="before")
    def parse_value(cls, data: dict) -> dict:
        # On subfilter validation does nothing
        if isinstance(data, dict):
            if data["oper"] in (Oper.ANY, Oper.HAS):
                typed_values = []
                for value in data["value"]:
                    value = to_string_or_numeric(value)
                    typed_values.append(value)
                data["value"] = typed_values

        return data

    def match(self, model) -> bool:
        return False


JsonApiQueryFilter.model_rebuild()


class JsonApiQueryModel(BaseModel):
    """JSON:API query validated model from flask pydantic.

    As request's query defines the fields to be returned, the filters and other parameters
    use this query to paginate the result or the not_found method if no result.
    """

    model_config = ConfigDict(frozen=True, str_strip_whitespace=True)

    fields: dict[str, list[str]] = Field(default_factory=dict)
    filter: list[str] | None = None
    include: set[str] = Field(default_factory=set)
    sort: str | None = None

    @computed_field  # type: ignore[misc]
    @cached_property
    def filters(self) -> t.Sequence[JsonApiQueryFilter]:
        """Returns an iterator over the filters from the query."""
        if isinstance(self.filter, t.Iterable):
            return [self._parse_filter(f) for f in self.filter]
        return []

    def get_fields(self, jsonapi_type) -> set[str] | None:
        """Returns the list of fields to dump for the given JSON API type."""
        if self.fields and jsonapi_type:
            values = self.fields.get(jsonapi_type)
            if values:
                return set(values)
        return None

    @property
    def query_pagination(self) -> JsonApiPagination:
        page = request.args.get("page[number]")
        size = request.args.get("page[size]")
        return JsonApiPagination(
            number=int(page) if page else 1,
            size=int(size) if size else 20,
        )

    def match(self, model: JsonApiBaseModel):
        """Checks if the model matches the query filters."""
        for filter in self.filters:
            if not filter.match(model):
                return False
        return True

    @staticmethod
    def no_content() -> Response:
        """Returns a no content response."""
        return Response(status=HTTPStatus.NO_CONTENT)

    def one(self, value: JsonApiBaseModel) -> TopLevel:
        """Returns a single JSON:API toplevel's data."""
        if value is None:
            raise NotFound()

        included: list[Resource] = []
        meta = {
            "count": 1,
        }
        return TopLevel(data=value.as_resource(included, self), included=included, meta=meta)  # noqa

    def one_or_none(self, value: JsonApiBaseModel) -> TopLevel:
        """Returns a single JSON:API toplevel's data."""
        if value is None:
            meta = {
                "count": 0,
            }
            return TopLevel(data=[], meta=meta)  # noqa

        return self.one(value)

    def paginate(
        self, values: list[Resource | JsonApiBaseModel], *, full_list: bool = True, total: int | None = None
    ) -> TopLevel:
        """Returns multi JSON:API toplevel's data from a value iterable.

        :param values: List for pagination.
        :param full_list: If true, the list is fully completed.
        :param total: The pagination total number of records if not a full list.
        """
        included: list[Resource] = []
        pagination = self.query_pagination
        total = len(values) if full_list else total
        if full_list:
            start = (pagination.number - 1) * pagination.size
            end = start + pagination.size
            values = values[start:end]
            has_next = end < len(values)
        else:
            values = values
            has_next = "tobedone"  # type: ignore

        meta = {
            "count": total,
            "pagination": {
                "page": pagination.number,
                "per_page": pagination.size,
                # "pages": page.pages,
                "has_next": has_next,
                "has_prev": pagination.number > 1,
            },
        }
        resources: list[Resource] = [v if isinstance(v, Resource) else v.as_resource(included, self) for v in values]
        return TopLevel.model_validate({"data": resources, "included": included, "meta": meta})

    @staticmethod
    def error(code: int, title: str, detail: str) -> tuple[TopLevel, int]:
        return (
            TopLevel(errors=[Error(id=str(code), title=title, detail=detail, status=str(code), code=str(code))]),
            code,
        )

    @staticmethod
    def not_found(detail: str | None = None) -> tuple[TopLevel, int]:
        error = NotFound(description=detail)
        return JsonApiQueryModel.error(error.code, error.name, error.get_description())  # type: ignore

    @model_validator(mode="before")
    def parse_filter_fields(cls, data: dict) -> dict:
        filter = []
        fields = {}
        for arg, value in data.items():
            if arg.startswith("filter["):
                filter.append(value)
            elif arg.startswith("fields["):
                groups = FIELDS_REGEXP.match(arg)
                if groups:
                    fields[groups[1]] = set(value.split(","))
                else:
                    msg = f"Wrong fields parameters: {arg}"
                    raise UnprocessableEntity(msg)

        if filter:
            data["filter"] = filter
        if fields:
            data["fields"] = fields

        return data

    @field_validator("include", mode="before")
    @classmethod
    def parse_include(cls, include: set[str] | list[str] | None) -> set[str]:
        if isinstance(include, set):
            return include
        return set(include[0].split(",")) if include else set()

    def _parse_filter(self, filter: str) -> JsonApiQueryFilter:
        if reg := COMPILED_FILTER_REGEXP.match(filter):
            oper = reg.group("oper")
            if oper:
                value = self._parse_value(reg)
                other = reg.group("other_attr")
                return JsonApiQueryFilter(
                    oper=Oper(oper), attr=reg.group("attr"), value=value, other_attr=other if other != "null" else None
                )
            oper = reg.group("list_oper")
            if oper:
                values = [v.strip() for v in reg.group("values").split(",")]
                return JsonApiQueryFilter(oper=Oper(oper), attr=reg.group("list_attr"), value=values)

        if reg := COMPILED_NOT_REGEXP.match(filter):
            filter_expr = reg.group("filter")
            return JsonApiQueryFilter(oper=Oper.NOT, subfilter=self._parse_filter(filter_expr))
        if reg := COMPILED_OR_AND_REGEXP.match(filter):
            filter1_expr = reg.group("filter1")
            filter2_expr = reg.group("filter2")
            return JsonApiQueryFilter(
                oper=Oper.NOT, subfilter=(self._parse_filter(filter1_expr), self._parse_filter(filter2_expr))
            )
        msg = f"Wrong filter parameters: {filter}"
        raise UnprocessableEntity(msg)

    def _parse_value(self, reg):
        numeric = reg.group("numeric")
        if numeric:
            value = float(numeric) if "." in numeric else int(numeric)
        else:
            value = reg.group("value")
            if value:
                value = value.strip("'")
                value = value.replace("''", "'")

        return value


class JsonApiBodyModel(BaseModel):
    model_config = ConfigDict(frozen=True, str_strip_whitespace=True)

    data: Resource | list[Resource] | None = None
    included: list[Resource] | None = Field(default_factory=list)
    debug: bool = False

    @property
    def attributes(self):
        if isinstance(self.data, list):
            return [d.attributes for d in self.data]
        return self.data.attributes if self.data else None

    @property
    def values(self):
        import warnings

        warnings.warn("Deprecated. Use 'attributes' instead.")
        return self.data



import json
import typing as t

from .jsonapi import Relationship, Resource, ResourceIdentifier, TopLevel


@t.overload
def extract_relationship(toplevel: dict | str, relationship: str) -> list[dict] | dict:
    """Extracts from dict or a JSON string."""


@t.overload
def extract_relationship(toplevel: TopLevel, relationship: Relationship | str) -> list[Resource] | Resource:
    """Extracts from a TopLevel model as a single resource."""


def extract_relationship(toplevel, relationship):
    """Returns the resource associated with the relationship if defined in included.

    :param toplevel: The toplevel jsonapi structure.
    :param relationship: The relationship to extract or the relationship path to extract.
    """
    tl = toplevel
    if isinstance(tl, str):
        tl = json.loads(tl)
    if isinstance(tl, dict):
        tl = TopLevel.model_validate(tl)
    if not tl.data:
        raise AttributeError("Wrong toplevel strucure: no data.")

    if isinstance(relationship, str):
        identifiers = get_relation_identifiers(tl, tl.data, relationship)
    elif isinstance(relationship, Relationship):
        identifiers = relationship.data
    else:
        raise AttributeError("Wrong relationship strucure.")

    if isinstance(identifiers, list):
        # Set in dictionary to avoid duplicate
        resources = {}
        for ident in identifiers:
            res = extract_from_resource_identifier(tl, ident)
            if isinstance(toplevel, str) or isinstance(toplevel, dict):
                resources[f"{res.type}{res.id}"] = res.model_dump()
            else:
                resources[f"{res.type}{res.id}"] = res
        return list(resources.values())

    res = extract_from_resource_identifier(tl, t.cast(ResourceIdentifier, identifiers))
    if isinstance(toplevel, str) or isinstance(toplevel, dict):
        return res.model_dump()
    return res


def extract_from_resource_identifier(toplevel: TopLevel, identifier: ResourceIdentifier) -> Resource:
    """Returns the resource associated with the resource identifier if defined in included.

    :param toplevel: The toplevel jsonapi structure.
    :param identifier: The resource identifier.
    """

    # Get all attributes values of the included resources
    if toplevel.included:
        for r in toplevel.included:
            if r.type == identifier.type and r.id == identifier.id:
                return r
    raise AttributeError(f"Cannot extract identifier {identifier} from {toplevel}")


def get_relation_identifiers(toplevel, data, relationship: str) -> ResourceIdentifier | list[ResourceIdentifier]:
    if isinstance(data, list):
        flat_list = []
        for d in data:
            rel = d.relationships[relationship].data
            flat_list.extend(rel) if isinstance(rel, list) else flat_list.append(rel)
        return flat_list

    if "." not in relationship:
        if relationship not in data.relationships:
            raise AttributeError(f"Cannot extract relationship {relationship} from {toplevel}")
        return data.relationships[relationship].data

    # Composed relationships
    relationship, other, *_ = relationship.split(".", 1)
    res = extract_relationship(toplevel, relationship)
    if isinstance(res, list):
        flat_list = []
        for r in res:
            rel = get_relation_identifiers(toplevel, r, other)
            flat_list.extend(rel) if isinstance(rel, list) else flat_list.append(rel)
        return flat_list
    else:
        return get_relation_identifiers(toplevel, res, other)


def flatten_resource(res: Resource | dict, *, more: dict | None = None) -> dict:
    """Returns the resource attributes with the 'id' added. Can add more data if needed."""
    more = more or {}
    if isinstance(res, Resource):
        return {"id": res.id, **res.attributes, **more}
    return {"id": res["id"], **res["attributes"], **more}


def to_bool(val: t.Any) -> bool:
    if isinstance(val, str):
        return val.lower() in ["true", "1", "yes", "y"]
    return bool(val)


def to_string_or_numeric(value: str) -> str | int | float:
    if value.startswith("'"):
        return value.strip("'")
    return float(value) if "." in value else int(value)



from __future__ import annotations

import types
import typing as t

from pydantic import AnyUrl, BaseModel, Field

from .jsonapi import Link, Relationship, Resource, ResourceIdentifier

if t.TYPE_CHECKING:
    from .dotnet import JsonApiQueryModel


class JsonApiBaseModel(BaseModel):
    """Base model for JSON:API resources."""

    jsonapi_type: t.ClassVar[str] = Field(frozen=True)

    @property
    def jsonapi_id(self) -> str:
        return getattr(self, "id", "")

    @property
    def links(self) -> dict[str, AnyUrl | Link | None] | None:
        return None

    def as_resource(
        self, included: list[Resource], query: JsonApiQueryModel, *, key_prefix: str | None = None
    ) -> Resource:
        jsonapi_id = self.jsonapi_id
        jsonapi_type = self.jsonapi_type
        relationships: dict[str, Relationship] = {}

        def _add_in_included(value, prefixed_key: str) -> ResourceIdentifier:
            """Adds the value in the included list if it is included in the query and returns a ResourceIdentifier."""
            if query.include and prefixed_key in query.include:
                to_be_incl: Resource = value.as_resource(included, query, key_prefix=prefixed_key)
                if len([r for r in included if r.type == to_be_incl.type and r.id == to_be_incl.id]) == 0:
                    included.append(to_be_incl)
            if isinstance(value, ResourceIdentifier):
                return value
            return ResourceIdentifier(type=value.jsonapi_type, id=value.jsonapi_id)

        fields = query.get_fields(jsonapi_type) if query else None

        # Creates relationships
        excluded_attributes = {"id", "type"}
        for key, f in self.__class__.model_fields.items():

            if fields is not None and key not in fields:
                continue

            value = getattr(self, key)
            prefixed_key = f"{key_prefix}.{key}" if key_prefix else key
            if issubtype(f.annotation, JsonApiBaseModel) and value is not None:  # type: ignore[arg-type]

                # Bypass if annotated as a dict
                if "as_attribute" in f.metadata:
                    continue

                if isinstance(value, list):
                    data = [_add_in_included(v, prefixed_key) for v in value]
                else:
                    data = _add_in_included(value, prefixed_key)
                relationships[key] = Relationship(data=data)  # noqa
                excluded_attributes.add(key)

        # Creates resource with attributes and relationships
        attributes = self.model_dump(include=fields, exclude=excluded_attributes)
        return Resource(
            id=jsonapi_id, type=jsonapi_type, attributes=attributes, relationships=relationships, links=self.links
        )


class JapydDictBaseModel(JsonApiBaseModel):
    """Base model for JSON:API resources."""

    @property
    def jsonapi_type(self) -> str:
        return getattr(self, "type", "")

    @jsonapi_type.setter
    def jsonapi_type(self, value: str) -> None:  # type: ignore
        assert False, "Cannot change the type of a resource."


T = t.TypeVar("T", bound="BaseModel")
UnionType = getattr(types, "UnionType", t.Union)


def issubtype(type_: t.Type, of_class: t.Generic[T]) -> T | None:  # type: ignore[valid-type]
    """Returns the subtype of a generic type if it is a subtype of the given class."""
    try:
        if issubclass(type(type_), types.GenericAlias) or issubclass(type_, t.Generic): # type: ignore[arg-type]
            type_ = t.get_args(type_)[0]
    except TypeError:
        pass

    origin = t.get_origin(type_) or type_
    if origin is t.Union or origin is UnionType:
        for type__ in t.get_args(type_):
            type__ = issubtype(type__, of_class)
            if type__:
                return type__
        return None

    try:
        return origin if issubclass(origin, of_class) else None  # type: ignore
    except TypeError:
        # Should check for ForwardRef
        return None



import typing as t

import requests
from pydantic import AnyUrl, BaseModel

from .jsonapi import Error, JsonApi, Link, Resource


class __TopLevel(BaseModel):
    errors: list[Error] | None = None
    meta: dict | None = None
    jsonapi: JsonApi | None = None
    links: dict[str, AnyUrl | Link | None] | None = None
    included: list[Resource] | None = None


#
# Duck typing TopLevel classes
#


class TopLevelSingle(__TopLevel):
    data: Resource


class TopLevelArray(__TopLevel):
    data: list[Resource]


class JapydClient:

    def __init__(self, url: str):
        self.url = url.rstrip("/")

    @t.overload
    def __call__(self, method: str, entry: str, /, id: str, **kwargs) -> TopLevelSingle: ...

    @t.overload
    def __call__(self, method: str, entry: str, /, **kwargs) -> TopLevelArray: ...

    def __call__(self, method: str, entry: str, /, id=None, **kwargs):
        if id is not None:
            resp = requests.request(method, f"{self.url.rstrip('/')}/{entry.strip('/')}/{id}", **kwargs)
            return TopLevelSingle.model_validate(resp.json())

        resp = requests.request(method, f"{self.url.rstrip('/')}/{entry}/{id}", **kwargs)
        return TopLevelArray.model_validate(resp.json())
